/* CC0 public domain */
#include <time.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/extensions/Xfixes.h>
#include <X11/extensions/XTest.h>

#define WHEEL_WAIT 16
#define TRIGGER_WAIT 12
#define BLINK_WAIT 8
#define SPEED_TIME_MAX 64
#define MOVE_MIN 1
#define MOVE_MAX 6

static KeySym
	trigger_on[]      = { XK_Shift_L, XK_Shift_R }
	, trigger_off[]   = { XK_q, XK_x, XK_Escape }
	, move_left[]     = { XK_g, XK_Left }
	, move_right[]    = { XK_j, XK_Right }
	, move_up[]       = { XK_y, XK_Up }
	, move_down[]     = { XK_h, XK_Down }
	, button_left[]   = { XK_c, XK_space }
	, button_middle[] = { XK_d }
	, button_right[]  = { XK_s }
	, wheel_up[]      = { XK_f }
	, wheel_down[]    = { XK_v };

static Display *display;
static char keymap[32];
static int enabled = False;

static int
key_pressed_single(KeySym keycode)
{
	KeyCode kc = XKeysymToKeycode(display, keycode);
	return (keymap[kc >> 3] & (1 << (kc & 7))) > 0;
}

#define KEY_PRESSED(X) (key_pressed_array((X), (sizeof (X) / sizeof (X)[0])))
static int
key_pressed_array(KeySym *keys, int len)
{
	for (int i = 0; i < len; i += 1) {
		if (key_pressed_single(keys[i])) return True;
	}
	return False;
}

static void
set_hiding(int is_hide)
{
	static int hiding = False;
	if (is_hide && !hiding) {
		XFixesHideCursor(display, DefaultRootWindow(display));
	}
	if (!is_hide && hiding) {
		XFixesShowCursor(display, DefaultRootWindow(display));
	}
	hiding = is_hide;
}

static void
scan_trigger(void)
{
	static int count_press = 0;
	static int prev_pressed = False;
	static int wating_time = 0;
	int pressed = False;
	int begin_press = False;

	/* off */
	if (enabled && KEY_PRESSED(trigger_off)) {
		enabled = False;
		XUngrabKeyboard(display, CurrentTime);
		set_hiding(False);
	}

	/* on */
	pressed = KEY_PRESSED(trigger_on);
	begin_press = pressed && !prev_pressed;
	prev_pressed = pressed;
	if (enabled) return;
	if (count_press > 0 && wating_time > 0) {
		wating_time -= 1;
		if (wating_time == 0) count_press = 0;
	}
	if (begin_press) {
		count_press += 1;
		wating_time = TRIGGER_WAIT;
	}
	else if (!pressed && wating_time > 0 && count_press == 2) {
		enabled = True;
		count_press = 0;
		XGrabKeyboard(display, DefaultRootWindow(display), True, GrabModeAsync, GrabModeAsync, CurrentTime);
	}
}

static void
control_mouse(void)
{
	static int prev_move_x = 0;
	static int prev_move_y = 0;
	static int speed_x = 0;
	static int speed_y = 0;
	static int anim = 0;
	static int prev_wheel_delta = 0;
	static int wheel_count_time = 0;

	int wheel_delta = 0, mx = 0, my = 0;
	int ax, ay;

	if (!enabled) {
		speed_x = 0;
		speed_y = 0;
		prev_move_x = 0;
		prev_move_y = 0;
		anim = 0;
		return;
	}

	/* click */
	XTestFakeButtonEvent(display, Button1, KEY_PRESSED(button_left), CurrentTime);
	XTestFakeButtonEvent(display, Button2, KEY_PRESSED(button_middle), CurrentTime);
	XTestFakeButtonEvent(display, Button3, KEY_PRESSED(button_right), CurrentTime);

	/* wheel */
	if (KEY_PRESSED(wheel_up)) wheel_delta -= 1;
	if (KEY_PRESSED(wheel_down)) wheel_delta += 1;
	if (prev_wheel_delta != wheel_delta) wheel_count_time = 0;
	if (wheel_delta != 0) {
		if (wheel_count_time < WHEEL_WAIT) wheel_count_time += 1;
	} else {
		wheel_count_time = 0;
	}
	if (wheel_count_time == 1 || wheel_count_time == WHEEL_WAIT) {
		int button = wheel_delta < 0 ? Button4 : Button5;
		XTestFakeButtonEvent(display, button, True, 1);
		XTestFakeButtonEvent(display, button, False, 1);
	}
	prev_wheel_delta = wheel_delta;

	if (KEY_PRESSED(move_left))  mx -= 1;
	if (KEY_PRESSED(move_right)) mx += 1;
	if (KEY_PRESSED(move_up))    my -= 1;
	if (KEY_PRESSED(move_down))  my += 1;

	/* idle */
	if (!mx && !my) {
		set_hiding(anim < BLINK_WAIT);
		anim = (anim + 1) % (BLINK_WAIT * 2);
		speed_x = 0;
		speed_y = 0;
		prev_move_x = 0;
		prev_move_y = 0;
		return;
	}
	anim = BLINK_WAIT;
	set_hiding(False);

	/* move */
	if (prev_move_x != mx) speed_x = 0;
	if (mx && speed_x < SPEED_TIME_MAX) speed_x += 1;
	if (prev_move_y != my) speed_y = 0;
	if (my && speed_y < SPEED_TIME_MAX) speed_y += 1;
	ax = MOVE_MIN + speed_x / (float)SPEED_TIME_MAX * (MOVE_MAX - MOVE_MIN);
	ay = MOVE_MIN + speed_y / (float)SPEED_TIME_MAX * (MOVE_MAX - MOVE_MIN);
	XWarpPointer(display, None, None, 0, 0, 0, 0, mx * ax, my * ay);
	XSync(display, False);
	prev_move_x = mx;
	prev_move_y = my;
}

static void
sleep()
{
	static const struct timespec ts = { .tv_sec = 0, .tv_nsec = 10000000 };
	nanosleep(&ts, NULL);
}

int
main(int argc, char *argv[])
{
	display = XOpenDisplay(NULL);
	if (!display) return 1;
	for (;;) {
		XQueryKeymap(display, keymap);
		scan_trigger();
		control_mouse();
		sleep();
	}
	/* unreachable */
	return 1;
}
